// cOlOr/C++
// e-mail: Franck.Barbier@FranckBarbier.com

#ifndef _Association_h
#define _Association_h

#ifdef _WIN32
#pragma warning (disable : 4786) // for debugging purposes with STL under Visual C++
#endif

#include <limits.h> // INT_MAX
#include <stddef.h> // NULL

#include <algorithm> // 'std::set_union'
#include <cassert>
#include <list>
#include <map>
#include <set>
#include <string>
#include <vector>

using namespace std;

template<typename T> class Role : public set<T*, less<T*> > {
private:
    int min;
    int max;
    typename set<T*, less<T*> >::iterator cursor;
    typename set<T*, less<T*> >::reverse_iterator reverse_cursor;

public:

    class RoleException {
    };

public:
    Role(T* t[] = NULL, int = 0, int = INT_MAX);
    Role(const vector<void*>&, int = 0, int = INT_MAX);
    Role(const Role<T>&);
    virtual ~Role();

public:
    virtual bool add(T*);
    virtual bool add(T&);
    vector<void*> as_vector() const;
    T* head();
    T* last();
    int multiplicity() const;
    T* next();
    T* previous();
    virtual bool remove(T*);
    virtual bool remove(T&);
    bool test(T*) const;
    bool test(T&) const;

public:
    const Role<T>& operator=(const Role<T>&);
};

template<typename T> Role<T>::Role(T* t[], int i, int j) {
    if (i >= 0 && j > 0 && i <= j) {
        min = i;
        max = j;
    } else throw RoleException();
    if (t) {
        int k = 0;
        while (t[k]) this->add(t[k++]);
        if (k < min) throw RoleException();
        if (k > max) throw RoleException();
    } else if (min) throw RoleException();
    cursor = this->begin();
    reverse_cursor = this->rbegin();
}

template<typename T> Role<T>::Role(const vector<void*>& v, int i, int j) {
    if (i >= 0 && j > 0 && i <= j) {
        min = i;
        max = j;
    } else throw RoleException();
    if (!v.empty()) {
        if (v.size() < min) throw RoleException();
        if (v.size() > max) throw RoleException();
        vector<void*>::const_iterator k = v.begin();
        for (; k != v.end(); k++) this->add((T*) * k);
    } else if (min) throw RoleException();
    cursor = this->begin();
    reverse_cursor = this->rbegin();
}

template<typename T> Role<T>::Role(const Role<T>& r) : set<T*, less<T*> >(r) {
    min = r.min;
    max = r.max;
    cursor = this->begin();
    reverse_cursor = this->rbegin();
}

template<typename T> Role<T>::~Role() {
}

template<typename T> bool Role<T>::add(T* t) {
    if (multiplicity() < max) return this->insert(t).second;
    return false;
}

template<typename T> bool Role<T>::add(T& t) {
    return add(&t);
}

template<typename T> vector<void*> Role<T>::as_vector() const {
    typename set<T*, less<T*> >::iterator i = this->begin();
    vector<void*> v;
    for (; i != this->end(); i++) v.push_back(*i);
    return v;
}

template<typename T> T* Role<T>::head() {
    if (this->empty()) return NULL;
    cursor = this->begin();
    return *cursor;
}

template<typename T> T* Role<T>::last() {
    if (this->empty()) return NULL;
    reverse_cursor = this->rbegin();
    return *reverse_cursor;
}

template<typename T> int Role<T>::multiplicity() const {
    return this->size();
}

template<typename T> T* Role<T>::next() {
    if (this->empty()) return NULL;
    if (++cursor == this->end()) {
        cursor = this->begin();
        return NULL;
    }
    return *cursor;
}

template<typename T> T* Role<T>::previous() {
    if (this->empty()) return NULL;
    if (++reverse_cursor == this->rend()) {
        reverse_cursor = this->rbegin();
        return NULL;
    }
    return *reverse_cursor;
}

template<typename T> bool Role<T>::remove(T* t) {
    if (this->multiplicity() > min && this->erase(t)) return true;
    return false;
}

template<typename T> bool Role<T>::remove(T& t) {
    return remove(&t);
}

template<typename T> bool Role<T>::test(T* t) const {
    if (this->count(t)) return true;
    return false;
}

template<typename T> bool Role<T>::test(T& t) const {
    return test(&t);
}

template<typename T> const Role<T>& Role<T>::operator=(const Role<T>& r) {
    if (this != &r) {
        this->set<T*, less<T*> >::operator=(r);
        min = r.min;
        max = r.max;
        cursor = this->begin();
        reverse_cursor = this->rbegin();
    }
    return *this;
}

template<typename Left, typename Right> class Association;

template<typename Left, typename Right> class _Association {
    friend class Association<Left, Right>;

private:
    map<Left*, Role<Right>, less<Left*> > all_rights;
    map<Right*, Role<Left>, less<Right*> > all_lefts;
    int min_Left;
    int max_Left;
    int min_Right;
    int max_Right;

private:
    list<Role<Right>* > garbage_Right;
    list<Role<Left>* > garbage_Left;

private:

    class _AssociationException {
    };

public: // for STL map
    _Association(int = 0, int = INT_MAX, int = 0, int = INT_MAX);
    ~_Association();

private:
    bool add(Left*, Right*);
    bool add(Left&, Right&);
    bool add(Right*, Left*);
    bool add(Right&, Left&);
    bool add(Left*, Role<Right>&);
    bool add(Left&, Role<Right>&);
    bool add(Right*, Role<Left>&);
    bool add(Right&, Role<Left>&);
    bool multiplicity_conformity();
    int multiplicity(Left*);
    int multiplicity(Left&);
    int multiplicity(Right*);
    int multiplicity(Right&);
    bool remove(Left*);
    bool remove(Left&);
    bool remove(Right*);
    bool remove(Right&);
    bool remove(Left*, Right*);
    bool remove(Left&, Right&);
    bool remove(Right*, Left*);
    bool remove(Right&, Left&);
    Role<Right> scan(Left*);
    Role<Right> scan(Left&);
    Role<Right> scan(Role<Left>&);
    Role<Left> scan(Right*);
    Role<Left> scan(Right&);
    Role<Left> scan(Role<Right>&);
    bool test(Left*);
    bool test(Left&);
    bool test(Right*);
    bool test(Right&);
    bool test(Left*, Right*);
    bool test(Left&, Right&);
    bool test(Right*, Left*);
    bool test(Right&, Left&);
};

template<typename Left, typename Right> _Association<Left, Right>::_Association(int i, int j, int k, int l) {
    if (i >= 0 && k >= 0 && j > 0 && l > 0 && i <= j && k <= l) {
        min_Left = i;
        max_Left = j;
        min_Right = k;
        max_Right = l;
    } else throw _AssociationException();
}

template<typename Left, typename Right> _Association<Left, Right>::~_Association() {
    typename list<Role<Right>* >::iterator i;
    for (i = garbage_Right.begin(); i != garbage_Right.end(); i++) delete (*i);
    typename list<Role<Left>* >::iterator j;
    for (j = garbage_Left.begin(); j != garbage_Left.end(); j++) delete (*j);
}

template<typename Left, typename Right> bool _Association<Left, Right>::add(Left* left, Right* right) {
    if (all_rights.find(left) == all_rights.end()) {
        // left does not exist
        if (all_lefts.find(right) == all_lefts.end()) {
            // left does not exist and right does not exist
            Role<Left>* lefts = new Role<Left>();
            garbage_Left.push_back(lefts);
            Role<Right>* rights = new Role<Right>();
            garbage_Right.push_back(rights);
            if (!lefts->add(left)) throw _AssociationException();
            if (!rights->add(right)) throw _AssociationException();
            all_lefts[right] = *lefts;
            all_rights[left] = *rights;
        } else {
            // left does not exist but right already exists
            assert(all_lefts[right].multiplicity() <= max_Left && !all_lefts[right].test(left));
            if (all_lefts[right].multiplicity() == max_Left) return false;
            Role<Right>* rights = new Role<Right>();
            garbage_Right.push_back(rights);
            if (!all_lefts[right].add(left)) throw _AssociationException();
            if (!rights->add(right)) throw _AssociationException();
            all_rights[left] = *rights;
        }
    } else {
        // left already exists
        assert(all_rights[left].multiplicity() <= max_Right);
        if (all_rights[left].multiplicity() == max_Right || all_rights[left].test(right)) return false;
        if (all_lefts.find(right) == all_lefts.end()) {
            // left already exists but right does not exist
            Role<Left>* lefts = new Role<Left>();
            garbage_Left.push_back(lefts);
            if (!all_rights[left].add(right)) throw _AssociationException();
            if (!lefts->add(left)) throw _AssociationException();
            all_lefts[right] = *lefts;
        } else {
            // left already exists and right already exists with no link between them
            assert(all_lefts[right].multiplicity() <= max_Left && !all_lefts[right].test(left));
            if (all_lefts[right].multiplicity() == max_Left) return false;
            if (!all_rights[left].add(right)) throw _AssociationException();
            if (!all_lefts[right].add(left)) throw _AssociationException();
        }
    }
    return true;
}

template<typename Left, typename Right> bool _Association<Left, Right>::add(Left& left, Right& right) {
    return add(&left, &right);
}

template<typename Left, typename Right> bool _Association<Left, Right>::add(Right* right, Left* left) {
    return add(left, right);
}

template<typename Left, typename Right> bool _Association<Left, Right>::add(Right& right, Left& left) {
    return add(&left, &right);
}

template<typename Left, typename Right> bool _Association<Left, Right>::add(Left* left, Role<Right>& rights) {
    if ((all_rights.find(left) == all_rights.end()) && rights.multiplicity() >= min_Right && rights.multiplicity() <= max_Right) {
        Right* right = rights.head();
        while (right) {
            if (all_lefts.find(right) != all_lefts.end()) {
                assert(all_lefts[right].multiplicity() <= max_Left);
                assert(!all_lefts[right].test(left));
                if (all_lefts[right].multiplicity() == max_Left) return false;
            }
            right = rights.next();
        }
        right = rights.head();
        while (right) {
            if (all_lefts.find(right) != all_lefts.end()) {
                assert(all_lefts[right].multiplicity() < max_Left);
                if (!all_lefts[right].add(left)) throw _AssociationException();
            } else {
                Role<Left>* lefts = new Role<Left>();
                garbage_Left.push_back(lefts);
                if (!lefts->add(left)) throw _AssociationException();
                all_lefts[right] = *lefts;
            }
            right = rights.next();
        }
        all_rights[left] = rights;
        return true;
    }
    return false;
}

template<typename Left, typename Right> bool _Association<Left, Right>::add(Left& left, Role<Right>& rights) {
    return add(&left, rights);
}

template<typename Left, typename Right> bool _Association<Left, Right>::add(Right* right, Role<Left>& lefts) {
    if ((all_lefts.find(right) == all_lefts.end()) && lefts.multiplicity() >= min_Left && lefts.multiplicity() <= max_Left) {
        Left* left = lefts.head();
        while (left) {
            if (all_rights.find(left) != all_rights.end()) {
                assert(all_rights[left].multiplicity() <= max_Right);
                assert(!all_rights[left].test(right));
                if (all_rights[left].multiplicity() == max_Right) return false;
            }
            left = lefts.next();
        }
        left = lefts.head();
        while (left) {
            if (all_rights.find(left) != all_rights.end()) {
                assert(all_rights[left].multiplicity() < max_Right);
                if (!all_rights[left].add(right)) throw _AssociationException();
            } else {
                Role<Right>* rights = new Role<Right>();
                garbage_Right.push_back(rights);
                if (!rights->add(right)) throw _AssociationException();
                all_rights[left] = *rights;
            }
            left = lefts.next();
        }
        all_lefts[right] = lefts;
        return true;
    }
    return false;
}

template<typename Left, typename Right> bool _Association<Left, Right>::add(Right& right, Role<Left>& lefts) {
    return add(&right, lefts);
}

template<typename Left, typename Right> bool _Association<Left, Right>::multiplicity_conformity() {
    typename map<Left*, Role<Right>, less<Left*> >::iterator i;
    for (i = all_rights.begin(); i != all_rights.end(); ++i) {
        assert((*i).second.multiplicity() <= max_Right);
        if ((*i).second.multiplicity() < min_Right) return false;
    }
    typename map<Right*, Role<Left>, less<Right*> >::iterator j;
    for (j = all_lefts.begin(); j != all_lefts.end(); ++j) {
        assert((*j).second.multiplicity() <= max_Left);
        if ((*j).second.multiplicity() < min_Left) return false;
    }
    return true;
}

template<typename Left, typename Right> int _Association<Left, Right>::multiplicity(Left* left) {
    if (all_rights.find(left) != all_rights.end()) return all_rights[left].multiplicity();
    return -1;
}

template<typename Left, typename Right> int _Association<Left, Right>::multiplicity(Left& left) {
    return multiplicity(&left);
}

template<typename Left, typename Right> int _Association<Left, Right>::multiplicity(Right* right) {
    if (all_lefts.find(right) != all_lefts.end()) return all_lefts[right].multiplicity();
    return -1;
}

template<typename Left, typename Right> int _Association<Left, Right>::multiplicity(Right& right) {
    return multiplicity(&right);
}

template<typename Left, typename Right> bool _Association<Left, Right>::remove(Left* left) {
    if (all_rights.find(left) != all_rights.end()) {
        Right *right = all_rights[left].head();
        while (right) {
            if (all_lefts.find(right) != all_lefts.end()) {
                assert(all_lefts[right].test(left));
                if (multiplicity_conformity() && all_lefts[right].multiplicity() == min_Left) return false;
            } else throw _AssociationException();
            right = all_rights[left].next();
        }
        right = all_rights[left].head();
        while (right) {
            if (all_lefts.find(right) != all_lefts.end()) {
                if (!all_lefts[right].remove(left)) throw _AssociationException();
            } else throw _AssociationException();
            right = all_rights[left].next();
        }
        all_rights.erase(left);
        return true;
    }
    return false;
}

template<typename Left, typename Right> bool _Association<Left, Right>::remove(Left& left) {
    return remove(&left);
}

template<typename Left, typename Right> bool _Association<Left, Right>::remove(Right* right) {
    if (all_lefts.find(right) != all_lefts.end()) {
        Left *left = all_lefts[right].head();
        while (left) {
            if (all_rights.find(left) != all_rights.end()) {
                assert(all_rights[left].test(right));
                if (multiplicity_conformity() && all_rights[left].multiplicity() == min_Right) return false;
            } else throw _AssociationException();
            left = all_lefts[right].next();
        }
        left = all_lefts[right].head();
        while (left) {
            if (all_rights.find(left) != all_rights.end()) {
                if (!all_rights[left].remove(right)) throw _AssociationException();
            } else throw _AssociationException();
            left = all_lefts[right].next();
        }
        all_lefts.erase(right);
        return true;
    }
    return false;
}

template<typename Left, typename Right> bool _Association<Left, Right>::remove(Right& right) {
    return remove(&right);
}

template<typename Left, typename Right> bool _Association<Left, Right>::remove(Left* left, Right* right) {
    if (all_rights.find(left) == all_rights.end()) return false;
    else {
        if (all_lefts.find(right) == all_lefts.end()) return false;
        else {
            if (multiplicity_conformity() && (all_rights[left].multiplicity() == min_Right || all_lefts[right].multiplicity() == min_Left)) return false;
            if (!all_rights[left].test(right)) {
                assert(!all_lefts[right].test(left));
                return false;
            }
        }
    }
    if (!all_rights[left].remove(right)) throw _AssociationException();
    if (!all_lefts[right].remove(left)) throw _AssociationException();
    return true;
}

template<typename Left, typename Right> bool _Association<Left, Right>::remove(Left& left, Right& right) {
    return remove(&left, &right);
}

template<typename Left, typename Right> bool _Association<Left, Right>::remove(Right* right, Left* left) {
    return remove(left, right);
}

template<typename Left, typename Right> bool _Association<Left, Right>::remove(Right& right, Left& left) {
    return remove(&right, &left);
}

template<typename Left, typename Right> Role<Right> _Association<Left, Right>::scan(Left* left) {
    Role<Right> rights;
    if (all_rights.count(left)) rights = all_rights[left];
    return rights;
}

template<typename Left, typename Right> Role<Right> _Association<Left, Right>::scan(Left& left) {
    return scan(&left);
}

template<typename Left, typename Right> Role<Right> _Association<Left, Right>::scan(Role<Left>& lefts) {
    Role<Right> result;
    Left* left = lefts.head();
    while (left) {
        Role<Right> rights = scan(left);
        (void) std::set_union(result.begin(), result.end(), rights.begin(), rights.end(), std::inserter(result, result.end()));
        left = lefts.next();
    }
    return result;
}

template<typename Left, typename Right> Role<Left> _Association<Left, Right>::scan(Right* right) {
    Role<Left> lefts;
    if (all_lefts.count(right)) lefts = all_lefts[right];
    return lefts;
}

template<typename Left, typename Right> Role<Left> _Association<Left, Right>::scan(Right& right) {
    return scan(&right);
}

template<typename Left, typename Right> Role<Left> _Association<Left, Right>::scan(Role<Right>& rights) {
    Role<Left> result;
    Right* right = rights.head();
    while (right) {
        Role<Left> lefts = scan(right);
        (void) std::set_union(result.begin(), result.end(), lefts.begin(), lefts.end(), std::inserter(result, result.end()));
        right = rights.next();
    }
    return result;
}

template<typename Left, typename Right> bool _Association<Left, Right>::test(Left* left) {
    if (all_rights.find(left) != all_rights.end()) return true;
    return false;
}

template<typename Left, typename Right> bool _Association<Left, Right>::test(Left& left) {
    return test(&left);
}

template<typename Left, typename Right> bool _Association<Left, Right>::test(Right* right) {
    if (all_lefts.find(right) != all_lefts.end()) return true;
    return false;
}

template<typename Left, typename Right> bool _Association<Left, Right>::test(Right& right) {
    return test(&right);
}

template<typename Left, typename Right> bool _Association<Left, Right>::test(Left* left, Right* right) {
    if (all_rights.find(left) != all_rights.end() && all_rights[left].test(right)) return true;
    return false;
}

template<typename Left, typename Right> bool _Association<Left, Right>::test(Left& left, Right& right) {
    return test(&left, &right);
}

template<typename Left, typename Right> bool _Association<Left, Right>::test(Right* right, Left* left) {
    if (all_lefts.find(right) != all_lefts.end() && all_lefts[right].test(left)) return true;
    return false;
}

template<typename Left, typename Right> bool _Association<Left, Right>::test(Right& right, Left& left) {
    return test(&right, &left);
}

template<typename Left, typename Right> class Association {
private:
    static map<string, _Association<Left, Right>, less<string> > _association;
    string _name;

public:

    class AssociationException {
    };

public:
    Association(int = 0, int = INT_MAX, int = 0, int = INT_MAX, string = "");
    ~Association();

public:
    bool add(Left*, Right*);
    bool add(Left&, Right&);
    bool add(Right*, Left*);
    bool add(Right&, Left&);
    bool add(Left*, Role<Right>&);
    bool add(Left&, Role<Right>&);
    bool add(Right*, Role<Left>&);
    bool add(Right&, Role<Left>&);
    bool multiplicity_conformity();
    int multiplicity(Left*);
    int multiplicity(Left&);
    int multiplicity(Right*);
    int multiplicity(Right&);
    bool remove(Left*);
    bool remove(Left&);
    bool remove(Right*);
    bool remove(Right&);
    bool remove(Left*, Right*);
    bool remove(Left&, Right&);
    bool remove(Right*, Left*);
    bool remove(Right&, Left&);
    Role<Right> scan(Left*);
    Role<Right> scan(Left&);
    Role<Right> scan(Role<Left>&);
    Role<Left> scan(Right*);
    Role<Left> scan(Right&);
    Role<Left> scan(Role<Right>&);
    bool test(Left*);
    bool test(Left&);
    bool test(Right*);
    bool test(Right&);
    bool test(Left*, Right*);
    bool test(Left&, Right&);
    bool test(Right*, Left*);
    bool test(Right&, Left&);
};

template<typename Left, typename Right> map<string, _Association<Left, Right>, less<string> > Association<Left, Right>::_association;

template<typename Left, typename Right> Association<Left, Right>::Association(int i, int j, int k, int l, string name) {
    _name = name;
    if (_association.find(_name) == _association.end()) {
        try {
            _association[_name] = _Association<Left, Right>(i, j, k, l);
        } catch (typename _Association<Left, Right>::_AssociationException) {
            throw AssociationException();
        }
    }
}

template<typename Left, typename Right> Association<Left, Right>::~Association() {
}

template<typename Left, typename Right> bool Association<Left, Right>::add(Left* left, Right* right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].add(left, right);
}

template<typename Left, typename Right> bool Association<Left, Right>::add(Left& left, Right& right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].add(left, right);
}

template<typename Left, typename Right> bool Association<Left, Right>::add(Right* right, Left* left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].add(right, left);
}

template<typename Left, typename Right> bool Association<Left, Right>::add(Right& right, Left& left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].add(right, left);
}

template<typename Left, typename Right> bool Association<Left, Right>::add(Left* left, Role<Right>& rights) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].add(left, rights);
}

template<typename Left, typename Right> bool Association<Left, Right>::add(Left& left, Role<Right>& rights) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].add(left, rights);
}

template<typename Left, typename Right> bool Association<Left, Right>::add(Right* right, Role<Left>& lefts) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].add(right, lefts);
}

template<typename Left, typename Right> bool Association<Left, Right>::add(Right& right, Role<Left>& lefts) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].add(right, lefts);
}

template<typename Left, typename Right> bool Association<Left, Right>::multiplicity_conformity() {
    assert(_association.find(_name) != _association.end());
    return _association[_name].multiplicity_conformity();
}

template<typename Left, typename Right> int Association<Left, Right>::multiplicity(Left* left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].multiplicity(left);
}

template<typename Left, typename Right> int Association<Left, Right>::multiplicity(Left& left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].multiplicity(left);
}

template<typename Left, typename Right> int Association<Left, Right>::multiplicity(Right* right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].multiplicity(right);
}

template<typename Left, typename Right> int Association<Left, Right>::multiplicity(Right& right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].multiplicity(right);
}

template<typename Left, typename Right> bool Association<Left, Right>::remove(Left* left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].remove(left);
}

template<typename Left, typename Right> bool Association<Left, Right>::remove(Left& left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].remove(left);
}

template<typename Left, typename Right> bool Association<Left, Right>::remove(Right* right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].remove(right);
}

template<typename Left, typename Right> bool Association<Left, Right>::remove(Right& right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].remove(right);
}

template<typename Left, typename Right> bool Association<Left, Right>::remove(Left* left, Right* right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].remove(left, right);
}

template<typename Left, typename Right> bool Association<Left, Right>::remove(Left& left, Right& right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].remove(left, right);
}

template<typename Left, typename Right> bool Association<Left, Right>::remove(Right* right, Left* left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].remove(right, left);
}

template<typename Left, typename Right> bool Association<Left, Right>::remove(Right& right, Left& left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].remove(right, left);
}

template<typename Left, typename Right> Role<Right> Association<Left, Right>::scan(Left* left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].scan(left);
}

template<typename Left, typename Right> Role<Right> Association<Left, Right>::scan(Left& left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].scan(left);
}

template<typename Left, typename Right> Role<Right> Association<Left, Right>::scan(Role<Left>& lefts) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].scan(lefts);
}

template<typename Left, typename Right> Role<Left> Association<Left, Right>::scan(Right* right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].scan(right);
}

template<typename Left, typename Right> Role<Left> Association<Left, Right>::scan(Right& right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].scan(right);
}

template<typename Left, typename Right> Role<Left> Association<Left, Right>::scan(Role<Right>& rights) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].scan(rights);
}

template<typename Left, typename Right> bool Association<Left, Right>::test(Left* left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].test(left);
}

template<typename Left, typename Right> bool Association<Left, Right>::test(Left& left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].test(left);
}

template<typename Left, typename Right> bool Association<Left, Right>::test(Right* right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].test(right);
}

template<typename Left, typename Right> bool Association<Left, Right>::test(Right& right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].test(right);
}

template<typename Left, typename Right> bool Association<Left, Right>::test(Left* left, Right* right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].test(left, right);
}

template<typename Left, typename Right> bool Association<Left, Right>::test(Left& left, Right& right) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].test(left, right);
}

template<typename Left, typename Right> bool Association<Left, Right>::test(Right* right, Left* left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].test(right, left);
}

template<typename Left, typename Right> bool Association<Left, Right>::test(Right& right, Left& left) {
    assert(_association.find(_name) != _association.end());
    return _association[_name].test(right, left);
}

#endif
